<?php

/**
 * A class representing a set of changes.
 */
class Changelog_nacoma_Model implements Countable {

	private $changes = array();
	private $raw_changes = array();
	private $since;
	private $for_object;
	private $attempt_full;

	private static $change_count = array(

	);

	/**
	 * Create an object for all changes in the specified object since $since.
	 * If the object isn't specified and the user isn't an admin, get all
	 * changes relevant to the current user. If the user is an admin, get all
	 * changes period.
	 */
	public function __construct ($since=null, $for_object=null, $attempt_full=false) {
		$this->since = $since;
		$this->for_object = $for_object;
		$this->attempt_full = $attempt_full;
		$this->load_changes();
	}

	/**
	 * Performs a query on its own since load_changes() is very much coupled
	 * with real world usage and this method is stupid.
	 *
	 * @return int
	 */
	public static function get_changes_count() {
		$user = op5Auth::instance()->get_user()->get_username();
		$sql = "select count(id) AS num_id from changelog where username =".nacoma_db::quote_string($user);
		$result = nacoma_db::exec_query($sql);
		$row = $result->fetch(PDO::FETCH_NUM);
		if($row && ($row[0] != 0)) {
			return $row[0];
		}
		return 0;
	}

	public static function get_type_count($type) {

		if (!isset(self::$change_count[$type])) {

			$user = op5Auth::instance()->get_user()->get_username();
			$sql = "select count(id) AS num_id from changelog where (object_type = " . nacoma_db::quote_string($type) . " AND username =".nacoma_db::quote_string($user) . ")";

			$result = nacoma_db::exec_query($sql);
			$row = $result->fetch(PDO::FETCH_NUM);

			if ($row && ($row[0] != 0)) self::$change_count[$type] = intval($row[0]);
			else self::$change_count[$type] = 0;

		}

		return self::$change_count[$type];


	}

	/**
	 * Return the raw database result as an array
	 */
	public function get_raw_changes() {
		return $this->raw_changes;
	}

	/**
	 * The function that actually retrieves the changes as specified to the
	 * constructor.
	 */
	public function load_changes () {

		$authuser = op5Auth::instance()->get_user();
		$rights = $authuser->get_auth_data();
		$user = $authuser->get_username();

		$query = 'SELECT * FROM changelog';

		if ($this->for_object != null) {
			$query .= ' WHERE oldname = '.nacoma_db::quote_string($this->for_object).' OR newname = '.nacoma_db::quote_string($this->for_object);
			if (is_numeric($this->for_object))
				$query .= ' OR object_id = '.nacoma_db::quote_string($this->for_object);
		}
		else if ($this->attempt_full && isset($rights['configuration_all'])) {
			$query .= ' WHERE 1=1 ';
		}
		else {
			$query .= ' WHERE (object_type, object_id) IN ' .
				'(SELECT object_type, object_id FROM changelog ' .
				'WHERE username = '.nacoma_db::quote_string($user).')';
		}

		if ($this->since != null)
			$query .= ' AND time > FROM_UNIXTIME('.nacoma_db::quote_string($this->since).')';

		$query .= ' ORDER BY id ASC';

		$result = nacoma_db::exec_query($query);
		$this->raw_changes = $result->fetchAll(PDO::FETCH_ASSOC);

		foreach ($this->raw_changes as $idx => $change) {
			$change = $this->convert_change($change);
			if (!$this->validate_change($change)) {
				unset($this->raw_changes[$idx]);
			}
			else {
				$this->changes[$idx] = $change;
			}
		}
	}

	/**
	 * Takes an array from PDO, and returns the object with it's states
	 * converted to a denormalized object, or false if there was an error.
	 */
	private function convert_change($change) {
		$class = $change['object_type'].'_nacoma_Model';
		$obj = new $class();
		try {
			if ($change['oldstate']) {
				$change['oldstate'] = convert_changelog_object($change['object_type'], $change['oldstate']);
				if ($change['oldstate'])
					$change['oldstate'] = $obj->denormalize($change['oldstate']);
			}
		} catch (NacomaRelationException $ex) {
		}

		try {
			if ($change['newstate']) {
				$change['newstate'] = convert_changelog_object($change['object_type'], $change['newstate']);
				if ($change['newstate'])
					$change['newstate'] = $obj->denormalize($change['newstate']);
			}
		} catch (NacomaRelationException $ex) {
			// If, say, a template was both added and deleted since the last save,
			// any users of the template that was also added and deleted won't be
			// saved. This is OK, because they were, after all, deleted, so we'll
			// clean everything back up shortly.
			return false;
		}
		return $change;
	}

	/**
	 * Returns true if a change is valid in this context, or false otherwise.
	 * If change only changes "secret" things, it is invalid
	 * If change is a deletion, it is invalid
	 *
	 * @param $change The change to validate
	 * @param $root_level If true, authorization checks will be disabled
	 */
	private function validate_change($change, $root_level=false) {
		global $DEBUG;
		$authuser = op5Auth::instance()->get_user();
		$rights = $authuser->get_auth_data();
		$user = $authuser->get_username();
		if ($change === false) {
			return false;
		}
		else if (!is_array($change)) {
			if ($DEBUG) {
				print "Doesn't look like a change: ";
				print_r($change);
			}
			return false;
		}
		else if (!$change['newstate'] && !$change['oldstate']) {
			if ($DEBUG) {
				print "Neither new state nor old state: ";
				print_r($change);
			}
			return false;
		}
		else if ($change['newstate'] === $change['oldstate'] && $change['oldname'] == $change['newname']) {
			if ($DEBUG) {
				print "Neither name nor config changed: ";
				print_r($change);
			}
			return false;
		}
		else if (!$root_level && !isset($rights['configuration_all']) && !$change['newstate'] && $change['username'] != $user) {
			// Someone else deleted this object. This often affects related
			// objects, in which case we loose if we don't include those.
			// It also makes pretty good sense to special case deletions.
			// (no debug text here, because this is a perfectly legal state)
			return false;
		}
		return true;
	}

	/**
	 * Return summary information about the changes contained.
	 * Summary is a one line string about the changes done.
	 */
	public function get_summary() {
		if (count($this) === 0)
			return 'There are no changes.';

		$ntype = array();
		foreach ($this->changes as $ary) {
			if (!isset($ntype[$ary['object_type']]))
				$ntype[$ary['object_type']] = array($ary['object_id'] => 1);
			else
				$ntype[$ary['object_type']][$ary['object_id']] = 1;
		}
		$str = 'There are '.count($this).' changes to ';
		$ntypes = count($ntype);
		$i = 1;
		foreach ($ntype as $type => $n) {
			if ($i == $ntypes)
				$str.= count($n) . ' ' . $type . ' objects';
			else if ($i + 1 == $ntypes)
				$str.= count($n) . ' ' . $type . ' objects and ';
			else
				$str.= count($n) . ' ' . $type . ' objects, ';
			$i++;
		}

		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$users = $this->get_users();
		$bystr = ' by ';
		if (($idx = array_search($user, $users)) !== false) {
			$bystr .= 'you and ';
			unset($users[$idx]);
		}

		if (count($users))
			$str .= $bystr.count($users).' other '.(count($users)==1?'user':'users').'.';
		else
			$str .= '.';

		return $str;
	}

	/**
	 * Returns a list of the users that made changes included in the changes
	 */
	public function get_users() {
		$users = array();
		foreach($this->changes as $change) {
			$users[$change['username']] = 1;
		}
		return array_keys($users);
	}

	/**
	 * Returns the names of all other users that has changes included in the
	 * changes.
	 */
	public function get_other_users() {
		$users = $this->get_users();
		$authuser = op5Auth::instance()->get_user();
		$current_user = $authuser->get_username();
		if (($idx = array_search($current_user, $users)) !== false)
			unset($users[$idx]);
		return $users;
	}

	/**
	 * Return the id of the changes included
	 */
	public function get_change_ids() {
		$ids = array();
		foreach ($this->changes as $change)
			$ids[$change['id']] = 1;
		return array_keys($ids);
	}

	/**
	 * Return array representing changes in semi-formal format
	 */
	public function get_list() {
		$changes = array();
		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$objchanges = array();
		$mergestring = null;
		foreach ($this->changes as $ary) {
			if (!$ary['oldstate'] && !$ary['newstate'])
				continue;
			else if ((!$ary['oldname'] && $ary['newname']) || !$ary['oldstate']) {
				$changes[] = array('type' => 'new',
					'object_type' => $ary['object_type'],
					'object_id' => $ary['object_id'],
					'name' => $ary['newname'],
					'user' => $ary['username'],
					'timestamp' => $ary['time']
				);
				$mergestring = '';
				continue;
			}
			else if (!$ary['newstate']) {
				$changes[] = array('type' => 'delete',
					'object_type' => $ary['object_type'],
					'object_id' => $ary['object_id'],
					'name' => $ary['oldname'],
					'user' => $ary['username'],
					'timestamp' => $ary['time']
				);
				$mergestring = '';
				continue;
			}

			// Two changes should be visually merged if they're chronologically
			// sequencial by the same user on the same object.
			// Detected beneath creation/deletion, as those should never be
			// merged with anything.
			if ($mergestring == $ary['username'].$ary['object_type'].$ary['object_id'])
				array_pop($changes);
			else
				$objchanges = array();
			$mergestring = $ary['username'].$ary['object_type'].$ary['object_id'];

			if ($ary['oldname'] != $ary['newname'])
				$objchanges[] = array('type' => 'rename',
					'old' => $ary['oldname'],
					'new' => $ary['newname']);

			if ($ary['oldstate'] === false || $ary['newstate'] === false) {
				$objchanges[] = array('type' => 'corruption',
					'name' => $ary['newname']);
			}
			else {
				foreach ($ary['oldstate'] as $oldkey => $oldval) {
					if (isset($ary['newstate'][$oldkey]) && $ary['newstate'][$oldkey] === $oldval) {
						unset($ary['newstate'][$oldkey]);
						continue;
					}
					else if (is_array($oldval) || (isset($ary['newstate'][$oldkey]) && is_array($ary['newstate'][$oldkey]))) {
						if (is_array($oldval) && isset($ary['newstate'][$oldkey]) && is_array($ary['newstate'][$oldkey])) {
							foreach ($oldval as $ok => $ov) {
								foreach ($ary['newstate'][$oldkey] as $nk => $nv) {
									if ($ov == $nv) {
										unset($oldval[$ok]);
										unset($ary['newstate'][$oldkey][$nk]);
									}
								}
							}
						}

						if (!empty($ary['newstate'][$oldkey]))
							$objchanges[] = array('type' => 'newmember',
								'new' => $ary['newstate'][$oldkey],
								'attribute' => $oldkey);
						if (!empty($oldval))
							$objchanges[] = array('type' => 'delmember',
								'old' => $oldval,
								'attribute' => $oldkey);
						unset($ary['newstate'][$oldkey]);
					}
					else if ($oldval && (!isset($ary['newstate'][$oldkey]) || !$ary['newstate'][$oldkey])) {
						$objchanges[] = array('type' => 'delete',
							'attribute' => $oldkey,
							'value' => $oldval);
						if (isset($ary['newstate'][$oldkey]))
							unset($ary['newstate'][$oldkey]);
					}
					else if ($oldval) {
						$objchanges[] = array('type' => 'change',
							'attribute' => $oldkey,
							'old' => $ary['oldstate'][$oldkey],
							'new' => $ary['newstate'][$oldkey]);
						unset($ary['newstate'][$oldkey]);
					}
				}

				foreach ($ary['newstate'] as $newkey => $newval)
					$objchanges[] = array('type' => 'new',
						'attribute' => $newkey,
						'value' => $newval);
			}

			if ($objchanges)
				$changes[] = array('type' => 'change',
					'object_type' => $ary['object_type'],
					'object_id' => $ary['object_id'],
					'name' => $ary['newname'],
					'user' => $ary['username'],
					'timestamp' => $ary['time'],
					'changes' => $objchanges);
		}

		return $changes;
	}
	/**
	 * Returns a detailed html list of the changes represented in here.
	 */
	public function get_html_list() {
		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$html = '';
		foreach ($this->get_list() as $entry) {
			$userstr = '';
			if ($entry['user'] !== $user) {
				$userstr = " by <b>".htmlspecialchars($entry['user'])."</b>";
			}
			switch($entry['type']) {
			 case 'new':
				 $html .= "<li>Created new ".htmlspecialchars($entry['object_type'])." object <b><a href=\"edit.php?obj_type=".htmlspecialchars($entry['object_type'])."&amp;obj_id=".htmlspecialchars($entry['object_id'])."\">".htmlspecialchars($entry['name'])."</a></b>$userstr at ".htmlspecialchars($entry['timestamp'])."</li>";
				break;
			 case 'delete':
				$html .= "<li>Deleted ".htmlspecialchars($entry['object_type'])." object <b>".htmlspecialchars($entry['name'])."</b>$userstr at ".htmlspecialchars($entry['timestamp'])."</li>";
				break;
			 case 'corruption':
				$innerhtml .= "<li>Unknown changes to <b><a href=\"edit.php?obj_type=".htmlspecialchars($entry['object_type'])."&amp;obj_id=".htmlspecialchars($entry['object_id'])."\">".htmlspecialchars($entry['name'])."</a></b> (can't tell you exactly what - the changelog seems corrupt).</li>";
				break;
			 case 'rename':
				$innerhtml .= "<li>Renamed object from <b>".htmlspecialchars($entry['old'])."</b> to <b><a href=\"edit.php?obj_type=".htmlspecialchars($entry['object_type'])."&amp;obj_id=".htmlspecialchars($entry['object_id'])."\">".htmlspecialchars($entry['new'])."</a></b>.</li>";
				break;
			 case 'change':
				$innerhtml = '';
				foreach ($entry['changes'] as $change) {
					switch($change['type']) {
					 case 'new':
						$innerhtml .= "<li>Added attribute <b>".htmlspecialchars($change['attribute'])."</b> with value <b>".htmlspecialchars($change['value'])."</b>.</li>";
						break;
					 case 'delete':
						$innerhtml .= "<li>Removed attribute <b>".htmlspecialchars($change['attribute'])."</b> (was <b>".htmlspecialchars($change['value'])."</b>).</li>";
						break;
					 case 'change':
						$innerhtml .= "<li>Changed attribute <b>".htmlspecialchars($change['attribute'])."</b> from <b>".htmlspecialchars($change['old'])."</b> to <b>".htmlspecialchars($change['new'])."</b>.</li>";
						break;
					 case 'newmember':
						foreach($change['new'] as $key => $value) {
							$change['new'][$key] = htmlspecialchars($value);
						}
						$innerhtml .= "<li>Added member(s) <b>".implode(', ', $change['new'])."</b> to attribute <b>".htmlspecialchars($change['attribute'])."</b>.</li>";
						break;
					 case 'delmember':
						foreach($change['old'] as $key => $value) {
							$change['old'][$key] = htmlspecialchars($value);
						}
						$innerhtml .= "<li>Removed member(s) <b>".implode(', ', $change['old'])."</b> from attribute <b>".htmlspecialchars($change['attribute'])."</b>.</li>";
						break;
					}
				}
				$html .= "<li>Changes in ".htmlspecialchars($entry['object_type'])." <b><a href=\"edit.php?obj_type=".htmlspecialchars($entry['object_type'])."&amp;obj_id=".htmlspecialchars($entry['object_id'])."\">".htmlspecialchars($entry['name'])."</a></b>$userstr at ".htmlspecialchars($entry['timestamp']).":<ul>\n{$innerhtml}\n</ul></li>";
				break;
			}
		}
		return '<ul class="changelisting">'.$html.'</ul>';
	}

	/**
	 * Returns a detailed plaintext list of the changes represented in here.
	 */
	public function get_text_list() {
		$plainstrings = array(
			'delete' => "Deleted %s object %s%s at %s\n",
			'rename' => "Renamed object from '%s' to '%s'.",
			'corruption' => "Unknown changes to %s.",
			'newmember' => "Added member(s) %s to attribute %s.",
			'delmember' => "Removed member(s) %s from attribute %s.",
			'newattr' => "Added attribute %s with value %s.",
			'delattr' => "Removed attribute %s (was %s).",
			'changeattr' => "Changed attribute %s from '%s' to '%s'.",
			'merge' => "Changes in %s object %s%s at %s:\n%s",
		);
		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$text = '';
		foreach ($this->get_list() as $entry) {
			 if ($entry['user'] !== $user)
				 $userstr = " by {$entry['user']}";
			 else
				 $userstr = '';
			switch($entry['type']) {
			 case 'new':
				$text .= "Created new {$entry['object_type']} object {$entry['name']}{$userstr} at {$entry['timestamp']}\n";
				break;
			 case 'delete':
				$text .= "Deleted {$entry['object_type']} object {$entry['name']}{$userstr} at {$entry['timestamp']}\n";
				break;
			 case 'corruption':
				$innertext .= "Unknown changes to {$entry['name']} (can't tell you exactly what - the changelog seems corrupt).\n";
				break;
			 case 'rename':
				$innertext .= "Renamed object from {$entry['old']} to {$entry['new']}.\n";
				break;
			 case 'change':
				$innertext = '';
				foreach ($entry['changes'] as $change) {
					switch($change['type']) {
					 case 'new':
						$innertext .= "\tAdded attribute {$change['attribute']} with value {$change['value']}.\n";
						break;
					 case 'delete':
						$innertext .= "\tRemoved attribute {$change['attribute']} (was {$change['value']}).\n";
						break;
					 case 'change':
						$innertext .= "\tChanged attribute {$change['attribute']} from {$change['old']} to {$change['new']}.\n";
						break;
					 case 'newmember':
						$innertext .= "\tAdded member(s) ".implode(', ', $change['new'])." to attribute {$change['attribute']}.\n";
						break;
					 case 'delmember':
						$innertext .= "\tRemoved member(s) ".implode(', ', $change['old'])." from attribute {$change['attribute']}.\n";
						break;
					}
				}
				$entry['changes'] = $innertext;
				$text .= "Changes in {$entry['object_type']} {$entry['name']}{$userstr} at {$entry['timestamp']}:\n{$entry['changes']}";
			}
		}
		return $text ? $text : "No changes\n";
	}

	/**
	 * Returns list of names, addresses and changes made for hosts.
	 */
	public function get_host_list() {
		$list = array();
		foreach ($this->changes as $ary) {
			if ($ary['object_type'] == "host") {
				if (!$ary['oldname'] || !$ary['oldstate']) {
					$list[] = array('name' => $ary['newname'], 'address' => isset($ary['newstate']['address'])?$ary['newstate']['address']:'Unknown address', 'change' => "Add");
				}
				else if (!$ary['newstate']) {
					$list[] = array('name' => $ary['oldname'], 'address' => isset($ary['oldstate']['address'])?$ary['newstate']['address']:'Unknown address', 'change' => "Delete");
				}
			}
		}
		return $list;
	}

	/**
	 * Implementation of the count interface: the number of "visually distinct"
	 * changes. Visually distinct changes are
	 */
	public function count() {
		$dates = 0;
		$last_change = 0;
		foreach($this->changes as $change) {
			$changestr = $change['object_id'].$change['object_type'].$change['username'];
			if (!$change['oldstate'] || !$change['newstate'] || $changestr != $last_change)
				$dates += 1;
			$last_change = $changestr;
		}
		return $dates;
	}

	/**
	 * Clear the changelog from changes visible to the current user
	 */
	public function delete_changes()
	{
		if (!count($this))
			return;
		$this->archive_changes();
		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$query = 'DELETE FROM changelog WHERE id IN ('.implode(',',$this->get_change_ids()).') OR username = '.nacoma_db::quote_string($user);
		nacoma_db::exec_query($query);
	}

	/**
	 * Save the selected changes into changelog_history to be able
	 * to track historical changes
	 */
	public function archive_changes()
	{
		if (!count($this))
			return;
		$authuser = op5Auth::instance()->get_user();
		$user = $authuser->get_username();
		$fields = array('object_type', 'object_name', 'oldname',
			'newname', 'oldstate', 'newstate', 'username', 'time');

		foreach ($this->changes as $ary) {
			$ary['object_name'] = $ary['newname'] ? $ary['newname'] : $ary['oldname'];
			if (isset($ary['oldstate'])) {
				$oldstate = serialize($ary['oldstate']);
				if ($oldstate !== false)
					$ary['oldstate'] = $oldstate;
			}
			if (isset($ary['newstate'])) {
				$newstate = serialize($ary['newstate']);
				if ($newstate !== false)
					$ary['newstate'] = $newstate;
			}
			$values = array();
			foreach ($fields as $key) {
				$values[$key] = $ary[$key];
			}

			$sql = "INSERT INTO changelog_history (".implode(',', $fields);
			$sql .= ") VALUES(:".implode(', :', array_keys($values)).')';
			$stmt = nacoma_db::prepare_statement($sql);
			$stmt->execute($values);
		}
	}

	/**
	 * Generate an object overlay that contains all differences between the
	 * current database state and the state that the user is allowed to export,
	 * as well as a list of affected files.
	 * We use this overlay on top of the actual database contents when writing
	 * objects to disk.
	 */
	public function undo_others_changes(&$overlay, &$out_changed_files)
	{
		// this contains the last state of deleted objects, as objects deleted
		// by other users must temporarily be resurrected
		$changed_files = array();
		$authuser = op5Auth::instance()->get_user();
		$rights = $authuser->get_auth_data();
		$user = $authuser->get_username();
		$queryall = 'SELECT * FROM changelog ORDER BY id DESC';
		# first, iterate over all objects backwards, to get to the oldest state
		# for all objects.
		$resultall = nacoma_db::exec_query($queryall);
		foreach($resultall->fetchAll(PDO::FETCH_ASSOC) as $ary) {
			if (!$this->validate_change($this->convert_change($ary), true))
				continue;
			$class = $ary['object_type'].'_nacoma_Model';
			$fromobj = convert_changelog_object($ary['object_type'], $ary['newstate']);
			$toobj = convert_changelog_object($ary['object_type'], $ary['oldstate']);
			if (isset($overlay[$ary['object_type']][$ary['object_id']]) && $overlay[$ary['object_type']][$ary['object_id']])
				$obj = $overlay[$ary['object_type']][$ary['object_id']];
			else if (isset($overlay[$ary['object_type']][$ary['object_id']]))
				continue;
			else
				$obj = new $class($ary['object_id']);
			$newobj = convert_object_state($obj, $fromobj, $toobj);
			if (!isset($overlay[$ary['object_type']]))
				$overlay[$ary['object_type']] = array();
			$overlay[$ary['object_type']][$ary['object_id']] = $newobj;
		}

		# then, iterate over them forwards, only updating and storing the chosen
		# objects
		foreach ($this->raw_changes as $ary) {
			$fromobj = convert_changelog_object($ary['object_type'], $ary['oldstate']);
			$toobj = convert_changelog_object($ary['object_type'], $ary['newstate']);
			$obj = $overlay[$ary['object_type']][$ary['object_id']];

			// object ID was reused from an object the user deleted after last
			// export
			if ($obj === false) {
				$class = $ary['object_type'].'_nacoma_Model';
				$obj = new $class();
				$obj->id = $ary['object_id'];
			}
			$obj = convert_object_state($obj, $fromobj, $toobj);

			// if the object has moved, we should update the old file
			if (isset($fromobj['file_id']) && $fromobj['file_id'])
				$changed_files[$fromobj['file_id']] = 1;
			if (isset($obj->obj['file_id']) && $obj->obj['file_id'])
				$changed_files[$obj->obj['file_id']] = 1;

			// Because half the time, references are stored in the other object
			// and I see no easy way to find out which direction the reference goes
			$references = get_object_reference_list($ary['object_type']);
			foreach (array($fromobj, $toobj) as $theobj) {
				if(!is_array($theobj))
					continue;
				foreach ($theobj as $key => $name) {
					if (!isset($references[$key]))
						continue;
					$res = false;

					$props = get_object_properties($ary['object_type']);
					switch ($props['var_index'][$key]) {
					 case 'mselect':
						$refs = array();
						if (!is_array($name))
							continue;
						if (empty($name))
							continue;
						foreach ($name as $n)
							$refs[] = nacoma_db::quote_string($n);
						$refs = '('.implode(',', $refs).')';
						$res = nacoma_db::exec_query("SELECT file_id FROM {$references[$key]} WHERE id IN $refs");
						break;
					 default:
						 continue;
					}
					if ($res == false)
						continue;
					foreach ($res->fetchAll(PDO::FETCH_ASSOC) as $row)
						$changed_files[$row['file_id']] = 1;
					unset($res);
				}
			}

			// if the object has been renamed, we need to work harder
			if (($ary['oldname'] && $ary['newname']) && $ary['oldname'] !== $ary['newname']) {
				// stupid nagios with your silly textual identifiers. We'll need to reexport
				// anything that references this object.
				$reverse_reference_types = get_object_reverse_reference_list($ary['object_type']);
				if (is_array($reverse_reference_types)) {
					foreach ($reverse_reference_types as $type => $keys) {
						foreach ($keys as $key) {
							$props = get_object_properties($type);
							switch ($props['var_index'][$key]) {
							 case 'select':
							 case 'hidden':
								$res = nacoma_db::exec_query("SELECT file_id FROM $type WHERE $key = {$ary['object_id']}", false);
								break;
							 case 'mselect':
								$jnc = get_junction_table_name($type, $key);
								$res = nacoma_db::exec_query("SELECT $type.file_id FROM $type INNER JOIN $jnc ON $jnc.$type = $type.id WHERE $jnc.".($key=="parents"?"parents":$ary['object_type'])." = {$ary['object_id']}");
								break;
							 default:
								 continue;
							}
							foreach ($res->fetchAll(PDO::FETCH_ASSOC) as $row)
								$changed_files[$row['file_id']] = 1;
							unset($res);
						}
					}
				}
			}
			// if the object is created or deleted, we can't look in the DB for
			// relations.
			// we could do this in a smart way, but because I'm daft, I'm just
			// gonna reexport everything I can find.
			else if (!$ary['oldname'] || !$ary['newname'] || !$ary['oldstate'] || !$ary['newstate']) {
				$reverse_reference_types = get_object_reverse_reference_list($ary['object_type']);
				if (is_array($reverse_reference_types)) {
					foreach ($reverse_reference_types as $type => $keys) {
						if ($type == 'FILE')
							continue;
						$res = nacoma_db::exec_query("SELECT file_id FROM $type", false);
						foreach ($res->fetchAll(PDO::FETCH_ASSOC) as $row)
							$changed_files[$row['file_id']] = 1;
						unset($res);
					}
				}
			}

			if ($ary['object_type'] === 'host' || $ary['object_type'] === 'hostgroup' && ($ary['oldname'] != $ary['newname'] || !$ary['oldname'] || !$ary['newname'] || !$ary['oldstate'] || !$ary['newstate'])) {
				// Stupid annoying special case! These types have grandchildren
				// which won't reference the object in the db, but on file.
				// I'm not even gonna bother, and just reexport all grandchild
				// types
				$grandchildren = array('serviceescalation', 'servicedependency');
				foreach ($grandchildren as $gc) {
					$res = nacoma_db::exec_query("SELECT file_id FROM $gc");
					foreach ($res->fetchAll(PDO::FETCH_ASSOC) as $row)
						$changed_files[$row['file_id']] = 1;
					unset($res);
				}
			}
			$overlay[$ary['object_type']][$ary['object_id']] = $obj;
		}

		// And now for another silly special case, brought to you by the moron
		// who created this mess in the first place!
		//
		// It's possible to have objects in overlay that aren't in one of the
		// $changed_files, so we must handle that inconsistency. We could
		// either add them to the changed file list (which is pointless), or
		// remove them from the overlay (as we now know we don't need it) and
		// have the export skip over them.
		foreach ($overlay as $type => $ary) {
			foreach ($ary as $idx => $obj) {
				if (!empty($obj) && !isset($changed_files[$obj->obj['file_id']]))
					unset($overlay[$type][$idx]);
			}
		}

		foreach ($changed_files as $filetype => $_) {
			$file = new FILE_nacoma_Model($filetype);
			if($file->obj) {
				$out_changed_files[] = $file->obj['file_name'];
			}
		}
	}

	/**
	 * Return true if the contact list needs to be recached, otherwise false
	 */
	public function is_recache_required() {
		// Relations between these object types must cause a recache
		// (I don't think new_{host,service} will ever make it here, but just
		// in case...)
		static $important_types = array(
			'contact',
			'contactgroup',
			'contact_template',
			'service',
			'new_service',
			'service_template',
			'host',
			'new_host',
			'host_template');
		foreach ($this->changes as $change) {
			$p = get_object_properties($change['object_type']);
			foreach (array('newstate', 'oldstate') as $state) {
				// we need to create/delete entries, as it's a new/deleted object
				if (!in_array($change['object_type'], $important_types))
					continue;

				if (!$change[$state] && in_array($change['object_type'], $important_types))
					return true;

				foreach ($change[$state] as $key => $_) {
					if (!isset($p['special'][$key]) || is_array($p['special'][$key]))
						continue;
					// we need to edit entries, as relations between important
					// objects have changed
					if (in_array($change['object_type'], $important_types) &&
					    in_array($p['special'][$key], $important_types))
						return true;
				}
			}
		}
		return false;
	}

	/**
	 * Fetches comments of old host/service.
	 * Prepares comments with new host/service names.
	 */
	public function handle_comments() {
		$cmds = array();
		// iterate and proceed if it's a rename operation.
		foreach ($this->changes as $ary) {
			if (($ary['oldname'] && $ary['newname']) && ($ary['oldname'] !== $ary['newname']) &&
				(($ary['object_type'] == "host") || ($ary['object_type'] == "service"))) {
				//Fetch comments if any with old host/service
				if ($ary['object_type'] == "host") {
					$filter_search = "Filter: host_name = ".$ary['oldname'];
				}
				else {
					$svcg = explode(';', $ary['oldname'], 2);
					if (count($svcg) == 1) {
						$svc = $svcg[0];
					}
					else {
						$svc = $svcg[1];
					}
					$filter_search = "Filter: service_description = ".$svc;
				}
				$ls = op5livestatus::instance();
				$comments = $ls->query('comments', array(
						$filter_search
					),
					array('comment', 'entry_time', 'author', 'host_name')
				);
				if ($comments[2] > 0) {
					foreach ($comments[1] as $key => $val) {
						if ($ary['object_type'] == "host") {
							$cmd = 'ADD_HOST_COMMENT;'.$ary['newname'];
						}
						else {
							$svcg = explode(';', $ary['newname'], 2);
							if (count($svcg) == 1) {
								$svc = $svcg[0];
							}
							else {
								$svc = $svcg[1];
							}
							$cmd = 'ADD_SVC_COMMENT;'.$val[3].';'.$svc;
						}
						$command = sprintf("[%d] %s;1;%s;%s", $val[1], $cmd, $val[2], $val[0]);
						array_push($cmds, $command);
					}
				}
			}
		}
		return $cmds;
	}

	/**
	 * Update preprocessed comments.
	 */
	public function update_comments($cmds) {
		$qh = op5queryhandler::instance();
		foreach ($cmds as $cmd) {
			try {
				$resp = $qh->call("command run", $cmd, '');
			} catch (op5queryhandler_Exception $e) {
				op5log::instance("nacoma")->log("error", "external command failed. Exception: " . $e->getMessage());
			}
		}
	}
}
